import random, config
from base import *
from engine import PygameEngine

class Bug(Creature):

    def __init__(self,syringe,geno = None,mask=0):
        Creature.__init__(self,geno)
        self.syringe = syringe
        self.mask = mask
        self.notify("created")

    def __repr__(self):
        output = ""
        for i in range(3,-1,-1):
            if (self.mask >> (i * 2)) & 3 == 3:
                output += " %s" % config.debug_names[i][self.pheno[1-i/2][1-i%2]]
        return output
        
    def test(self,phevo):
        if self.mask != 0 and phevo.geno & self.mask == self.geno & self.mask:
            return True
        return False

class Syringe(Agent):
    def __init__(self, player, bugs = None):
        Agent.__init__(self)
        self.player = player
        if not bugs:
            self.bugs = []
        else:
            self.bugs = bugs
        self.mask = config.mask

    def spray(self,target):
        if len(self.bugs) > 0:
            bug = self.bugs.pop(0)
            self.notify("sprayed",target,bug)
            self.debug("sprayed")
            
            #self.fill()
#        if len(self.bugs) == 0:
#            self.notify("emptied")

    def fill(self,phevo = None):
        #if not phevo:
        phevo = random.choice(self.player.pop.well.keys())
        bug = Bug(self,phevo.geno,self.mask)
        bug.mask = bug.create_random_mask(False)
        self.bugs.append(bug)
        self.notify("filled",bug)
        phevo.promote()

class Phevo(Creature):

    def __init__(self, geno = None, population = None, index = None):
        Creature.__init__(self,geno)
        self.population = population
        self.index = index
        self.infected = False
        self.dead = False
        self.promoted = False
        self.neighbors = {}
        self.bug = None

    def __repr__(self):
        return "[%d] %s %s %s %s" % (self.index,\
                                config.debug_names[3][self.pheno[0][0]], \
                                config.debug_names[2][self.pheno[0][1]], \
                                config.debug_names[1][self.pheno[1][0]], \
                                config.debug_names[0][self.pheno[1][1]])

    def breed(self, mate, allow_mutants = True):
        mask = self.create_random_mask()
        offspring = Phevo(mask & self.geno | ~mask & mate.geno)
        if allow_mutants and random.randint(1,config.mutation_rate) == 1:
            offspring.geno ^= 1 << random.randint(0,7)
            self.debug("Mutation!")
        return offspring

    def check(self):
        if self.infected:
            self.cough()
            if random.randint(1,config.cure_rate) == 1:
                self.cure()
            if random.randint(1,config.kill_rate) == 1:
                self.die()

    def connect_neighbor(self,neighbor, dir=0):
        if neighbor:
            if not self.neighbors.has_key(neighbor):
                self.neighbors[neighbor] = dir
            if not neighbor.neighbors.has_key(self):
                neighbor.neighbors[self] = (dir + config.max_neighbors/2) % config.max_neighbors

    def disconnect_neighbor(self,neighbor):
        if self.neighbors.has_key(neighbor):
            self.neighbors.pop(neighbor)
        if neighbor.neighbors.has_key(self):
            neighbor.neighbors.pop(self)

    def disconnect_neighbors(self):
        for neighbor in self.neighbors.keys():
            if neighbor.neighbors.has_key(self):
                neighbor.neighbors.pop(self)
        self.neighbors = {}

    def infect(self,bug):
        self.debug("%s infected with %s" % (self,bug))
        self.infected = True
        self.bug = bug
        self.notify("infected")        

    def cure(self):
        self.infected = False
        self.notify("cured")

    def die(self):
        self.infected = False
        self.dead = True
        self.notify("died")
        self.disconnect_neighbors()
        self.bug = None
        self.population = None
        
    def promote(self):
        self.infected = False
        self.dead = False
        self.promoted = True
        self.notify("promoted")
        self.disconnect_neighbors()
        self.bug = None
        self.population = None
        
    def on_bug_attacked(self,bug,phevo):
        if self==phevo:
            self.infect(bug)

    def on_bug_cured(self,bug,phevo):
        if self==phevo:
            self.cure()

    def on_phevopop_check_requested(self,population):
        if population == self.population:
            self.check()

    def cough(self):
        self.notify("coughed")
        for phevo in self.neighbors.keys():
            if not phevo.infected \
               and not phevo.dead \
               and not phevo.bug == self.bug \
               and self.bug.test(phevo) \
               and random.randint(1,config.infect_rate) == 1: #immunity
                phevo.infect(self.bug)
        
class PhevoPop(Agent):
    def __init__(self, player, phevos = None):
        Agent.__init__(self)
        self.player = player
        if phevos:
            self.phevos = phevos
        else:
            self.phevos = []
        self.sick = {}
        self.well = {}

    def populate(self, phevos = None, generations = 6):
        generations -= 1
        if not phevos:
            phevos = []
            phevos.append(Phevo())
            phevos.append(Phevo())
        while generations > 0:
            new_pop = []
            for m in range(2):
                for x in range(len(phevos)):
                    new_pop.append(phevos[x].breed(phevos[x + (x % 2) * -2 + 1]))
            phevos = new_pop
            return self.populate(phevos, generations)
        if generations == 0:
            self.phevos = phevos
            random.shuffle(self.phevos)
            for index,phevo in enumerate(self.phevos):
                phevo.index = index
                phevo.population = self
                self.well[phevo] = None
                #phevo.subscribe(self)
                phevo.notify("created")


    def connect(self):
        height, width = (8,8)
        for y in range(height):
            for x in range(width):
                index = x + width * y
                """
                Directions:
                  0
                3 + 1
                  2
                """
                if x % width != 0:
                    self.phevos[index].connect_neighbor(self.phevos[index - 1],1)
                #if x == width:
                #    self.phevos[index].connect_neighbor(self.phevos[-1 * width])
                    
                if y > 0:
                    self.phevos[index].connect_neighbor(self.phevos[index - 1 * width],2)
                #if y == height:
                #    self.phevos[index].connect_neighbor(self.phevos[x])
                """                    
                if x % width != 0 and y > 0:
                    subjects[-1].subscribe(subjects[-1 * width - 2])
                if x % width != width-1 and y > 0:
                    subjects[-1].subscribe(subjects[-1 * width - 0])
                """

    def update(self):
        if self.player.ticks % config.population_update_rate == 0:
            self.notify("check_requested")

    def remove_phevo(self,phevo):
        if phevo.population == self:
            if self.sick.has_key(phevo):
                self.sick.pop(phevo)
            if self.well.has_key(phevo):
                self.well.pop(phevo)
            self.phevos[phevo.index] = None
            if len(self.well) + len(self.sick) == 0:
                self.notify("empty")
            
    def on_syringe_sprayed(self,syringe,target,bug):
        if target == self and len(self.well) > 0:
            self.debug("invaded")
            attempts = 0
            while attempts < len(self.well):
                phevo = random.choice(self.well.keys())
                if bug.test(phevo):
                    bug.notify("attacked",phevo)
                    break
                attempts += 1
                #else:
                #    bug.notify("released",phevo)

    def on_phevo_created(self,phevo):
        self.notify("populated",phevo)

    def on_phevo_infected(self,phevo):
        if self.well.has_key(phevo):
            self.well.pop(phevo)
        self.sick[phevo] = None

    def on_phevo_cured(self,phevo):
        if self.sick.has_key(phevo):
            self.sick.pop(phevo)
        self.well[phevo] = None

    def on_phevo_died(self,phevo):
        self.remove_phevo(phevo)
            
    def on_phevo_promoted(self,phevo):
        if phevo.population == self:
            index = phevo.index
            self.debug("Player %d Phevo %d promoted" % (phevo.population.player.id + 1, index))
            self.remove_phevo(phevo)
            phevo=random.choice(self.well.keys()).breed(random.choice(self.well.keys()))
            self.phevos[index] = phevo
            phevo.index = index
            phevo.population = self
            self.well[phevo] = None
            phevo.notify("created")
            
            x = phevo.index % config.grid_width
            y = phevo.index / config.grid_height
            
            self.debug("Connecting to (clockwise) %d %d %d %d" % (phevo.index - 1,phevo.index + 1,phevo.index - 1 * config.grid_width,phevo.index + 1 * config.grid_width))
            if x != 0:
                phevo.connect_neighbor(self.phevos[phevo.index - 1],1)
            if x < config.grid_width - 1:
                phevo.connect_neighbor(self.phevos[phevo.index + 1],1)
            if y > 0:
                phevo.connect_neighbor(self.phevos[phevo.index - 1 * config.grid_width],2)
            if y < config.grid_height - 1:
                phevo.connect_neighbor(self.phevos[phevo.index + 1 * config.grid_width],2)

            

class Player(Agent):
    def __init__(self,app,id,local = True, AI = True):
        Agent.__init__(self)
        self.app = app
        self.id = id
        self.pop = PhevoPop(self)
        self.syringe = Syringe(self)
        self.local = local
        self.AI = AI
        
        self.opponents = {}
        
        self.pop.populate()
        self.pop.connect()

        self.ticks = 0L
        
        #for i in range(2):
        #    self.syringe.fill()

    def add_opponent(self,player):
        if not self.opponents.has_key(player):
            self.opponents[player] = 0
        if not player.opponents.has_key(self):
            player.opponents[self] = 0
        
    def sub_opponent(self,player):
        if self.opponents.has_key(player):
            self.opponents.pop(player)
        if player.opponents.has_key(self):
            player.opponents.pop(self)

    def on_pygameengine_ticked(self,engine):
        self.ticks +=1
        self.pop.update()

    def on_pygameengine_spray_requested(self,engine,id):
        if id == self.id:
            self.debug("spraying")
            #this is where you'll need to add the choice of phevo
            self.syringe.fill()
            for opponent in self.opponents.keys():
                self.syringe.spray(opponent.pop)
        
    def on_phevopop_empty(self,pop):
        if self == pop.player:
            self.notify("defeated")
    #def on_syringe_emptied(self,syringe):
    #    if syringe == self.syringe:
    #        syringe.fill()

class App(Agent):
    def __init__(self,engine):
        Agent.__init__(self)
    	self.engine = engine
        self.players = range(config.players)
        for player_id in range(len(self.players)):
            self.players[player_id] = Player(self,player_id,AI = False)
            if player_id > 0:
                self.players[player_id].add_opponent(self.players[player_id - 1])
            self.players[player_id].syringe.fill()

    def start(self):

        #main loop
        while not(self.engine.finished):
            self.engine.tick()

    def stop(self):
        self.engine.stop()
        
if __name__ == "__main__":
    engine = PygameEngine(config.width,config.height)

    app = App(engine)
    try:
        app.start()
    finally:
        app.stop()    
